
#include "QCamView3.h"

#include "qcam_driver.h"
 
#include "VideoView.h"
#include "VideoWindow.h"

#define bmsgVideoFrame 'vidf'

VideoView::VideoView(BRect r,ulong size,int bpp)
:BView(r,"VideoView",B_FOLLOW_ALL,B_WILL_DRAW)
{
	qcam	qc;
	
	m_qcamDevice=open("/dev/quickcam",O_RDONLY|O_BINARY);
	
	// need this later
	m_sem=create_sem(1,"");
	
	ioctl(m_qcamDevice,QC_GET,&qc);
	
	qc.brightness=BRIGHTNESS;
	qc.contrast=CONTRAST;
	qc.whitebalance=WHITEBALANCE;
	
	CreateBitmaps(size,qc);
	
	switch(bpp)
	{
		case 4:
			qc.bpp=QC_4BIT_PIXEL;
			break;
			
		case 6:
			qc.bpp=QC_6BIT_PIXEL;
			break;
	}
	
	ioctl(m_qcamDevice,QC_SET,&qc);
}

void VideoView::CreateBitmaps(ulong size, qcam &qc)
{
	switch(size)
	{
		case QCAM_80x60:
			qc.factor = QC_FACTOR_41;
			m_numPix = 80*60;
			m_bitmap1 = new BBitmap(BRect(0,0,79,59),B_COLOR_8_BIT,FALSE);
			m_bitmap2 = new BBitmap(BRect(0,0,79,59),B_COLOR_8_BIT,FALSE);
			break;
			
		case QCAM_160x120:
			qc.factor = QC_FACTOR_21;
			m_numPix = 160*120;
			m_bitmap1 = new BBitmap(BRect(0,0,159,119),B_COLOR_8_BIT,FALSE);
			m_bitmap2 = new BBitmap(BRect(0,0,159,119),B_COLOR_8_BIT,FALSE);
			break;
			
		case QCAM_320x240:
			qc.factor = QC_FACTOR_11;
			m_numPix = 320*240;
			m_bitmap1 = new BBitmap(BRect(0,0,319,239),B_COLOR_8_BIT,FALSE);
			m_bitmap2 = new BBitmap(BRect(0,0,319,239),B_COLOR_8_BIT,FALSE);
			break;
	}

	m_bitmap=m_bitmap1;
	m_nextBitmap=m_bitmap2;
}

VideoView::~VideoView()
{
	close(m_qcamDevice);
	delete m_bitmap;
	delete m_nextBitmap;
	delete_sem(m_sem);
}

void VideoView::AttachedToWindow()
{
	inherited::AttachedToWindow();
	Window()->PostMessage(new BMessage(bmsgVideoFrame),this);
}

void VideoView::DoFrame()
{	
	// because we may be resizing the bitmaps at any time
	// protect this with a semaphore
	acquire_sem(m_sem);
	
	read(m_qcamDevice,m_nextBitmap->Bits(),m_numPix);
	BBitmap *bm = m_bitmap;
	m_bitmap = m_nextBitmap;
	m_nextBitmap = bm;
	
	// no longer async, so that we properly hang on to the semaphore
	DrawBitmap(m_bitmap,m_bitmap->Bounds(),Bounds());
	
	release_sem(m_sem);
}

void VideoView::ResizeBitmaps(ulong msize)
{
	qcam	qc;
	ulong	size;
	
	// because we are always trying to draw frames
	// protect this with a semaphore
	acquire_sem(m_sem);
	
	delete m_bitmap1;
	delete m_bitmap2;
	
	ioctl(m_qcamDevice,QC_GET,&qc);
	
	switch(msize)
	{
		case mSMALL:
			size=QCAM_80x60;
			break;
			
		case mMEDIUM:
			size=QCAM_160x120;
			break;
			
		case mLARGE:
			size=QCAM_320x240;
			break;
	}
	
	CreateBitmaps(size,qc);
	
	Window()->ResizeTo(m_bitmap->Bounds().Width(),m_bitmap->Bounds().Height());
	
	ioctl(m_qcamDevice,QC_SET,&qc);
	
	release_sem(m_sem);
}

void VideoView::AdjustBitDepth(ulong bpp)
{
	qcam	qc;
	
	ioctl(m_qcamDevice,QC_GET,&qc);
	switch(bpp)
	{
		case m4BITS:
			qc.bpp = QC_4BIT_PIXEL;
			break;
			
		case m6BITS:
			qc.bpp = QC_6BIT_PIXEL;
			break;
	}
	ioctl(m_qcamDevice,QC_SET,&qc);
}

void VideoView::SetWhiteBalance(unsigned char newValue)
{
	qcam	qc;
	
	ioctl(m_qcamDevice,QC_GET,&qc);
	qc.whitebalance=newValue;
	ioctl(m_qcamDevice,QC_SET,&qc);
}

void VideoView::SetContrast(unsigned char newValue)
{
	qcam	qc;
	
	ioctl(m_qcamDevice,QC_GET,&qc);
	qc.contrast=newValue;
	ioctl(m_qcamDevice,QC_SET,&qc);
}

void VideoView::SetBrightness(unsigned char newValue)
{
	qcam	qc;
	
	ioctl(m_qcamDevice,QC_GET,&qc);
	qc.brightness=newValue;
	ioctl(m_qcamDevice,QC_SET,&qc);
}

void VideoView::MessageReceived(BMessage *msg)
{
	switch(msg->what)
	{
		case bmsgVideoFrame:
			DoFrame();
			Window()->PostMessage(new BMessage(bmsgVideoFrame),this);
			break;

//		case mSMALL:
//		case mMEDIUM:
//		case mLARGE:
//			ResizeBitmaps(msg->what);
//			break;
//			
//		case m4BITS:
//		case m6BITS:
//			AdjustBitDepth(msg->what);
//			break;
				
		case B_MESSAGE_NOT_UNDERSTOOD:
//			 destination didn't understand the 'PPNT' message type. The
//			 only other one is a 'ref'.
			{
				BMessage	*original;
				bool		ref_ok=FALSE;
				long		i=0;
				
				original=msg->Previous();
				//ASSERT(original);
				if (original->what=='PPNT')
				{
					// try sending the data in a file instead...
					BMessage *new_msg=convert_to_ref_message(original);
					msg->SendReply(new_msg);
				}
			}
			break;
	}
}

void VideoView::Draw(BRect updateRect)
{
	DrawBitmap(m_bitmap,m_bitmap->Bounds(),Bounds());
}

BFile *VideoView::write_bitmap(BMessage *msg,char *fname,BDirectory *dir)
{
	char		*bits;
	long		bl;
	long		temp;
	BFile		*a_file;
	BRect 		tmpBound;
	BRect 		bnd;
	color_space	tc;	
	long		e0;

	a_file = new BFile();
	
	e0 = dir->Create(fname, a_file);
	if (dir->Error()) {
		e0 = dir->GetFile(fname, a_file);
		e0 = a_file->Open(B_READ_WRITE);
	}

	long	tmp_long;

	bnd = msg->FindRect("bounds");
	tmpBound = bnd;
	a_file->Open(B_READ_WRITE);	
	tmp_long = tmpBound.left;
	a_file->Write((char *)&tmp_long, 4);
	tmp_long = tmpBound.top;
	a_file->Write((char *)&tmp_long, 4);
	tmp_long = tmpBound.right;
	a_file->Write((char *)&tmp_long, 4);
	tmp_long = tmpBound.bottom;
	a_file->Write((char *)&tmp_long, 4);

	tc = (color_space) msg->FindLong("color_space");
	temp = tc;
	a_file->Write((char *)&(temp), 4);
	bl = msg->FindLong("bits_length");
	a_file->Write((char *)&(bl), 4);
	bits = (char *) msg->FindData("bits", B_RAW_TYPE, &temp);
	a_file->Write((char *)bits, bl);
	a_file->Close();
	a_file->SetTypeAndApp('BITM', 'PPNT');
	return(a_file);
}

BMessage *VideoView::convert_to_ref_message(BMessage *msg)
{
	BDirectory	root_dir;
	BDirectory	tmp_dir;
	BFile		*a_file;
	BVolume		tmp_volume;

	tmp_volume = boot_volume();
	tmp_volume.GetRootDirectory(&root_dir);
	if (root_dir.GetDirectory("tmp", &tmp_dir) != B_NO_ERROR)
		root_dir.Create("tmp", &tmp_dir);
	//ASSERT(root_dir.Error() == B_NO_ERROR);

	a_file = write_bitmap(msg,CLIP_FILENAME, &tmp_dir);
	msg = new BMessage(B_SIMPLE_DATA /* B_REFS_RECEIVED */);
	msg->AddRef("refs", a_file->Record()->Ref());
	delete a_file;

	return msg;
}

BMessage *VideoView::create_clip_message(BBitmap *bm)
{
	BMessage *msg = new BMessage('PPNT');

	msg->AddRect("bounds", bm->Bounds());
	msg->AddLong("color_space", bm->ColorSpace());
	msg->AddLong("bits_length", bm->BitsLength());
	msg->AddData("bits", B_RAW_TYPE, bm->Bits(), bm->BitsLength());
	
	return msg;
}

void VideoView::drag_selection(long h,long v,BPoint mouseDown)
{
	BBitmap			*selection_bm;
	BRect			selection_bound;
	BRect			new_dst;
	BRect			old_dst;
	long			h0, v0;
	long			h1, v1;
	ulong			buttons;
	long			dh, dv;
	BView			*tmp_view;
	BRect			vert, horiz;
	BRect			view_bound;
	BMessage		*msg;

//	clip(&h, &v);

	dv = dh = 0;

	view_bound = Bounds();

	selection_bound.top=0;
	selection_bound.bottom=selection.bottom-selection.top;
	selection_bound.left=0;
	selection_bound.right=selection.right-selection.left;

	selection_bm=new BBitmap(selection_bound,m_bitmap->ColorSpace(),TRUE);
	selection_bm->Lock();
	selection_bm->AddChild(tmp_view = new BView(selection_bound, "", B_FOLLOW_ALL, B_WILL_DRAW));
	tmp_view->DrawBitmap(m_bitmap, selection, selection_bound);
	selection_bm->Unlock();

	msg = create_clip_message(selection_bm);

	dh = mouseDown.x;
	dv = mouseDown.y;
	dv -= selection.top;
	dh -= selection.left;

	//PRINT(("dragging messaage (%.4s)\n", (char *) &(msg->what)));
	DragMessage(msg, selection_bm, BPoint(dh, dv));
}

void VideoView::MouseDown(BPoint where)
{
	long	x=where.x;
	long	y=where.y;

	selection.top=0;
	selection.left=0;
	selection.bottom=m_bitmap->Bounds().Height();
	selection.right=m_bitmap->Bounds().Width();

	Window()->Activate(TRUE);
	
	drag_selection(x,y,where);
}
